import sys
from collections import defaultdict
from pprint import pprint
from typing import Dict, List

import numpy as np
import torch
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification, AutoTokenizer

from a2t.topic_classification.base import TopicClassifier, np_softmax


class _NLITopicClassifier(TopicClassifier):

    def __init__(self, pretrained_model, topics, *args, use_cuda=True,
                 query_phrase="The domain of the sentence is about", entailment_position=2, 
                 half=False, verbose=True, **kwargs):
        super().__init__(pretrained_model, topics, use_cuda=use_cuda, verbose=verbose, half=half)
        self.query_phrase = query_phrase
        self.ent_pos = entailment_position

    def _initialize(self, pretrained_model):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model)
        self.model = AutoModelForSequenceClassification.from_pretrained(pretrained_model)

    def _run_batch(self, batch):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True)
            input_ids = torch.tensor(input_ids['input_ids']).to(self.device)
            output = self.model(input_ids)[0][:, self.ent_pos].view(input_ids.shape[0] // len(self.topics), -1)
            output = output.detach().cpu().numpy()

        return output

    def __call__(self, contexts: List[str], batch_size: int = 1):
        if not isinstance(contexts, list):
            contexts = [contexts]

        batch, outputs = [], []
        for i, context in tqdm(enumerate(contexts), total=len(contexts)):
            sentences = [f"{context} {self.tokenizer.sep_token} {self.query_phrase} \"{topic}\"." for topic in
                        self.topics]
            batch.extend(sentences)

            if (i + 1) % batch_size == 0:
                output = self._run_batch(batch)
                outputs.append(output)
                batch = []

        if len(batch) > 0:
            output = self._run_batch(batch)
            outputs.append(output)

        outputs = np.vstack(outputs)

        return outputs


class NLITopicClassifier(_NLITopicClassifier):

    def __init__(self, pretrained_model: str, topics: List[str], *args, **kwargs):
        super(NLITopicClassifier, self).__init__(pretrained_model, topics, *args, **kwargs)

    def __call__(self, contexts: List[str], batch_size: int = 1):
        outputs = super().__call__(contexts=contexts, batch_size=batch_size)
        outputs = np_softmax(outputs)

        return outputs


class NLITopicClassifierWithMappingHead(_NLITopicClassifier):

    def __init__(self, pretrained_model: str, topics: List[str], topic_mapping: Dict[str, str], *args, **kwargs):
        self.new_topics = list(topic_mapping.keys())
        self.target_topics = topics
        self.new_topics2id = {t: i for i, t in enumerate(self.new_topics)}
        self.mapping = defaultdict(list)
        for key, value in topic_mapping.items():
            self.mapping[value].append(self.new_topics2id[key])

        super().__init__(pretrained_model, self.new_topics, *args, **kwargs)

    def __call__(self, contexts, batch_size=1):
        outputs = super().__call__(contexts, batch_size)
        outputs = np.hstack([np.max(outputs[:, self.mapping[topic]], axis=-1, keepdims=True)
                             for topic in self.target_topics])
        outputs = np_softmax(outputs)

        return outputs


if __name__ == "__main__":

    if len(sys.argv) < 2:
        print('Usage:\tpython3 get_topics.py topics.txt input_file.txt\n\tpython3 get_topics.py topics.txt < '
              'input_file.txt')
        exit(1)

    with open(sys.argv[1], 'rt') as f:
        topics = [topic.rstrip().replace('_', ' ') for topic in f]

    input_stream = open(sys.argv[2], 'rt') if len(sys.argv) == 3 else sys.stdin

    clf = NLITopicClassifier('roberta-large-mnli', topics=topics)

    for line in input_stream:
        line = line.rstrip()
        output_probs = clf(line)[0]
        topic_dist = sorted(list(zip(output_probs, topics)), reverse=True)
        print(line)
        pprint(topic_dist)
        print()



import argparse
import json
import os
from pprint import pprint

import numpy as np
from sklearn.metrics import precision_recall_fscore_support

from a2t.topic_classification.mlm import MLMTopicClassifier
from a2t.topic_classification.mnli import NLITopicClassifier, NLITopicClassifierWithMappingHead
from a2t.topic_classification.nsp import NSPTopicClassifier

CLASSIFIERS = {
    'mnli': NLITopicClassifier,
    'nsp': NSPTopicClassifier,
    'mlm': MLMTopicClassifier,
    'mnli-mapping': NLITopicClassifierWithMappingHead
}


def top_k_accuracy(output, labels, k=5):
    preds = np.argsort(output)[:, ::-1][:, :k]
    return sum(l in p for l, p in zip(labels, preds)) / len(labels)


parser = argparse.ArgumentParser(prog='run_evaluation', description="Run a evaluation for each configuration.")
parser.add_argument('dataset', type=str,
                    help="Dataset file.")
parser.add_argument('topics', type=str,
                    help="Topics or classes file.")
parser.add_argument('--config', type=str, dest='config',
                    help='Configuration file for the experiment.')

args = parser.parse_args()

with open(args.topics, 'rt') as f:
    topics = [topic.rstrip().replace('_', ' ') for topic in f]

topic2id = {topic: i for i, topic in enumerate(topics)}

with open(args.dataset, 'rt') as f:
    contexts, labels = [], []
    for line in f:
        _, label, context = line.strip().split('\t')
        contexts.append(context)
        labels.append(topic2id[label])

labels = np.array(labels)

with open(args.config, 'rt') as f:
    config = json.load(f)

for configuration in config:
    os.makedirs(f"experiments/{configuration['name']}", exist_ok=True)
    classifier = CLASSIFIERS[configuration['classification_model']](topics=topics, **configuration)
    output = classifier(contexts, batch_size=configuration['batch_size'])
    np.save(f"experiments/{configuration['name']}/output.npy", output)
    np.save(f"experiments/{configuration['name']}/labels.npy", labels)
    pre, rec, f1, _ = precision_recall_fscore_support(labels, np.argmax(output, -1), average='weighted')
    configuration['precision'] = pre
    configuration['recall'] = rec
    configuration['f1-score'] = f1
    configuration['top-1'] = top_k_accuracy(output, labels, k=1)
    configuration['top-3'] = top_k_accuracy(output, labels, k=3)
    configuration['top-5'] = top_k_accuracy(output, labels, k=5)
    configuration['topk-curve'] = [top_k_accuracy(output, labels, k=i) for i in range(len(topics))]
    pprint(configuration)

with open(args.config, 'wt') as f:
    json.dump(config, f, indent=4)



import argparse
import gzip
import json
import os

import numpy as np

from a2t.topic_classification.mlm import MLMTopicClassifier
from a2t.topic_classification.mnli import NLITopicClassifier, NLITopicClassifierWithMappingHead
from a2t.topic_classification.nsp import NSPTopicClassifier

CLASSIFIERS = {
    'mnli': NLITopicClassifier,
    'nsp': NSPTopicClassifier,
    'mlm': MLMTopicClassifier,
    'mnli-mapping': NLITopicClassifierWithMappingHead
}

parser = argparse.ArgumentParser(prog='annotate_wordnet', description="Annotate WordNet glosses with domain information.")
parser.add_argument('dataset', type=str,
                    help="Dataset file.")
parser.add_argument('topics', type=str,
                    help="Topics or classes file.")
parser.add_argument('--config', type=str, dest='config',
                    help='Configuration file for the experiment.')

args = parser.parse_args()

with open(args.topics, 'rt') as f:
    topics = [topic.rstrip().replace('_', ' ') for topic in f]

topic2id = {topic:i for i, topic in enumerate(topics)}

senses, glosses = [], []
with gzip.open(args.dataset, 'rt') as f:
    for line in f:
        row = line.split()
        idx, gloss = row[0], " ".join(row[1:])
        gloss = gloss.replace('\t', ' ')
        senses.append(idx)
        glosses.append(gloss)

with open(args.config, 'rt') as f:
    config = json.load(f)

for configuration in config:
    os.makedirs(f"experiments/{configuration['name']}", exist_ok=True)
    classifier = CLASSIFIERS[configuration['classification_model']](topics=topics, **configuration)
    output = classifier(glosses, batch_size=configuration['batch_size'])
    # Normalize the output to avoid low confidence level due to large ammount of labels
    output = (output - output.min()) / (output.max() - output.min())
    labels = [topics[i] for i in np.argmax(output, -1)]
    confidences = [x for x in np.max(output, -1)]

    with gzip.open(f"experiments/{configuration['name']}/wordnet.annotations.tsv.gz", 'wt') as out_f:
        for sense, label, confidence in zip(senses, labels, confidences):
            out_f.write(f"{sense}\t{label}\t{confidence}\n")



from .mlm import MLMTopicClassifier
from .mnli import NLITopicClassifierWithMappingHead, NLITopicClassifier
from .nsp import NSPTopicClassifier

__all__ = ['NLITopicClassifierWithMappingHead',
           'NLITopicClassifier',
           'MLMTopicClassifier',
           'NSPTopicClassifier']



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="a2t",
    version="0.1",
    author="Oscar Sainz",
    author_email="osainz006@ehu.eus",
    description="Ask2Transformers is a library for zero-shot classification based on Transformers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/osainz59/Ask2Transformers",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'transformers',
        'tqdm',
        'torch'
    ]
)



import sys
from pprint import pprint

import numpy as np
import torch
from tqdm import tqdm
from transformers import BertForNextSentencePrediction, AutoTokenizer

from a2t.topic_classification.base import TopicClassifier


class NSPTopicClassifier(TopicClassifier):

    def __init__(self, pretrained_model, topics, *args, use_cuda=True, query_phrase="Topic or domain about",
                 positive_position=1, half=False, **kwargs):
        super().__init__(pretrained_model, topics, use_cuda=use_cuda, half=half)
        self.query_phrase = query_phrase
        self.cls_pos = positive_position

    def _initialize(self, pretrained_model):
        self.tokenizer = AutoTokenizer.from_pretrained(pretrained_model, use_fast=True)
        self.model = BertForNextSentencePrediction.from_pretrained(pretrained_model)

    def _run_batch(self, batch):
        with torch.no_grad():
            input_ids = self.tokenizer.batch_encode_plus(batch, padding=True)
            input_ids = torch.tensor(input_ids['input_ids']).to(self.device)
            output = self.model(input_ids)[0][:, self.cls_pos].view(len(batch) // len(self.topics), -1)
            output = torch.softmax(output, dim=-1).detach().cpu().numpy()

        return output

    def __call__(self, contexts, batch_size=1):
        if not isinstance(contexts, list):
            contexts = [contexts]

        batch, outputs = [], []
        for i, context in tqdm(enumerate(contexts), total=len(contexts)):
            sentences = [f"{context} {self.tokenizer.sep_token} {self.query_phrase} \"{topic}\"." for topic in
                         self.topics]
            batch.extend(sentences)

            if (i + 1) % batch_size == 0:
                output = self._run_batch(batch)
                outputs.append(output)
                batch = []

        if len(batch) > 0:
            output = self._run_batch(batch)
            outputs.append(output)

        outputs = np.vstack(outputs)

        return outputs


if __name__ == "__main__":

    if len(sys.argv) < 2:
        print(
            'Usage:\tpython3 get_topics.py topics.txt input_file.txt\n\tpython3 get_topics.py topics.txt < '
            'input_file.txt')
        exit(1)

    with open(sys.argv[1], 'rt') as f:
        topics = [topic.rstrip().replace('_', ' ') for topic in f]

    input_stream = open(sys.argv[2], 'rt') if len(sys.argv) == 3 else sys.stdin

    clf = NSPTopicClassifier('bert-large-uncased', topics=topics)

    for line in input_stream:
        line = line.rstrip()
        output_probs = clf(line)[0]
        topic_dist = sorted(list(zip(output_probs, topics)), reverse=True)
        print(line)
        pprint(topic_dist)
        print()
